#include "GoWindow.h"

#include <Autolock.h>

#include "besupport/BThread.h"

#include "NetworkWindow.h"

using namespace muscle;

extern BMessageTransceiverThread gMuscle;

void
go_event_callback( Go * go, int stone, XY xy, void * _data )
{
	GoWindow * win = (GoWindow*)_data;
	
	win->GoEvent( stone, xy );
}

GoWindow::GoWindow( NetworkWindow * owner, Go * go, const char * name, const char * remote_addr, const char * local_addr, const char * session )
:	BWindow( 
		BRect(50,50,350,350), string(string("Go - ") + string(name)).c_str(), B_TITLED_WINDOW, B_ASYNCHRONOUS_CONTROLS
	),
	fOwner(owner),
	fGo( go ),
	fLocalAddress( local_addr ),
	fSession( session ),
	fPassed( false )
{
	fGo->AddListener( go_event_callback, this );
	
	BRect string_rect = Bounds();
	string_rect.bottom = 25;
	BRect score_rect(string_rect);
	BRect pass_rect(string_rect);
	string_rect.right -= 120;
	score_rect.left = string_rect.right+1;
	score_rect.right -= 50;
	pass_rect.left = score_rect.right+1;
	
	BRect go_rect = Bounds();
	go_rect.top = string_rect.bottom+1;
	
	fGoView = new GoView(go_rect, fGo);
	
	fGoView->SetEnabled(false);
	
	AddChild( fGoView );
	
	fMessageView = new BStringView( 
		string_rect, 
		"MessageView", 
		"A message..", 
		B_FOLLOW_LEFT_RIGHT|B_FOLLOW_TOP
	);
	
	AddChild( fMessageView );
	
	fScoreView = new BStringView( 
		score_rect, 
		"ScoreView", 
		"W: 0 B: 0", 
		B_FOLLOW_RIGHT|B_FOLLOW_TOP
	);
	
	AddChild( fScoreView );
	
	fPassButton = new BButton(
		pass_rect,
		"passButton",
		"Pass",
		new BMessage('Pass'),
		B_FOLLOW_RIGHT|B_FOLLOW_TOP
	);
	
	fPassButton->SetEnabled(false);
	
	AddChild( fPassButton );
	
	// connect network
	if ( !remote_addr )
	{
		fMessageView->SetText("Waiting for opponent..");
		printf("Waiting for opponent [go_session '%s']\n", fSession.c_str() );
//		MessageRef msg = GetMessageFromPool('Publ');
//		gMuscle.SendMessageToSessions(msg);
	} else
	{
		fGoView->fColor = Go::Black;
		
		fMessageView->SetText("Connecting to opponent.. ");
		
		printf("Connecting to opponent [go_session '%s']\n", fSession.c_str() );
		
		fRemoteAddress = remote_addr;
		
		// send join message to opponent
		MessageRef msg = GetMessageFromPool('Join');
		msg()->AddString(PR_NAME_KEYS, fRemoteAddress.c_str() );
		msg()->AddString("address", fLocalAddress.c_str() );
		msg()->AddString("go_session", fSession.c_str() );
		
		printf("Message to send:\n");
		msg()->PrintToStream();
		
		gMuscle.SendMessageToSessions(msg);
	}
}

GoWindow::~GoWindow()
{
	if ( fRemoteAddress == "" )
	{
		// remove game publish
		MessageRef remove_msg = GetMessageFromPool(PR_COMMAND_REMOVEDATA);
		remove_msg()->AddString(PR_NAME_KEYS, string("Go/Publish/" + fSession).c_str() );
		gMuscle.SendMessageToSessions(remove_msg);
	} else
	{
		// tell remote side we're quiting
		MessageRef quit_msg = GetMessageFromPool('Quit');
		quit_msg()->AddString(PR_NAME_KEYS, fRemoteAddress.c_str() );
		quit_msg()->AddString("go_session", fSession.c_str() );
		gMuscle.SendMessageToSessions(quit_msg);
	}
	
	BMessage msg('Remo');
	msg.AddString("go_session", fSession.c_str() );
	BMessenger(fOwner).SendMessage(&msg);
	
	delete fGo;
}

void
GoWindow::MessageReceived( BMessage * msg )
{
	switch ( msg->what )
	{
		case 'Sugg':
		{
			int32 x=-1, y=-1;
			
			msg->FindInt32("x", &x);
			msg->FindInt32("y", &y);
			
			MessageRef nmsg = GetMessageFromPool('Sugg');
			
			nmsg()->AddString(PR_NAME_KEYS, fRemoteAddress.c_str() );
			nmsg()->AddString("go_session", fSession.c_str() );
			nmsg()->AddInt32("x", x );
			nmsg()->AddInt32("y", y );
			
			gMuscle.SendMessageToSessions(nmsg);
		}	break;
		
		case 'Pass':
		{
			// Player passed
			int32 x=-1, y=-1;
			
			msg->FindInt32("x", &x);
			msg->FindInt32("y", &y);
			
			MessageRef nmsg = GetMessageFromPool('Pass');
			
			nmsg()->AddString(PR_NAME_KEYS, fRemoteAddress.c_str() );
			nmsg()->AddString("go_session", fSession.c_str() );
			
			gMuscle.SendMessageToSessions(nmsg);
			
			//
			fGoView->SetEnabled(false);
			fPassButton->SetEnabled(false);
			
			if ( !fPassed )
			{
				fPassed = true;
				fMessageView->SetText("You passed. Opponent's turn.");
			} else {
				fMessageView->SetText("Game over.");
			}
		}	break;
		
		default:
			BWindow::MessageReceived(msg);
			break;
	}
}

void
GoWindow::HandleGoMessage( Message * msg )
{
	BAutolock locker(this);
	
	switch ( msg->what )
	{
		case 'Busy':
		{
			printf("Got a 'busy' reply\n");
			fMessageView->SetText("Opponent already in game, fail!");
		}	break;
		
		case 'Conn':
		{ // connection successful
			printf("Connected\n");
			fMessageView->SetText("Waiting for opponent's move..");
		}	break;
		
		case 'Join':
		{
			if ( fRemoteAddress == "" )
			{ // someone has joined, nice.
				printf("Got a join\n");
				
				const char * remote_addr = NULL;
				
				if ( msg->FindString("address", &remote_addr) != B_OK )
				{
					printf("Failed to get remote address from message\n");
					msg->PrintToStream();
					return;
				}
				
				fRemoteAddress = remote_addr;
				
				// send 'Conn' message to let them know we've connected
				MessageRef reply = GetMessageFromPool('Conn');
				reply()->AddString("go_session", fSession.c_str() );
				gMuscle.SendMessageToSessions(reply);
				
				// remove game publish
				MessageRef remove_msg = GetMessageFromPool(PR_COMMAND_REMOVEDATA);
				remove_msg()->AddString(PR_NAME_KEYS, string("Go/Publish/" + fSession).c_str() );
				gMuscle.SendMessageToSessions(remove_msg);
				
				// Now it's our turn to place a stone
				fMessageView->SetText("Your turn.");
				
				fGoView->SetEnabled(true);
				fPassButton->SetEnabled(true);
			} else
			{ // already playing, send failure message
				printf("Got a join - busy\n");
				
				// send 'Conn' message to let them know we've connected
				MessageRef reply = GetMessageFromPool('Busy');
				reply()->AddString("go_session", fSession.c_str() );
				gMuscle.SendMessageToSessions(reply);
			}
		}	break;
		
		case 'Move':
		{
			fPassed = false;
			
			// perform remote move
			Go::Stone col;
			if ( fGoView->fColor == Go::White )
				col = Go::Black;
			else
				col = Go::White;
			
			fGo->UpdateValidPlacements( col );
			
			int32 x=-1,y=-1;
			msg->FindInt32("x", &x);
			msg->FindInt32("y", &y);
			
			fGo->PlaceStone(col,x,y);
			
			// set up for own move
			fGo->UpdateValidPlacements( fGoView->fColor );
			fGoView->SetEnabled(true);
			fPassButton->SetEnabled(true);
			fMessageView->SetText("Your move.");
			UpdateScore();
		}	break;
		
		case 'Sugg':
		{
			// Opponent suggested move
			int32 x=-1,y=-1;
			msg->FindInt32("x", &x);
			msg->FindInt32("y", &y);
			
			fGoView->SetSuggestion( XY(x,y) );
		}	break;
		
		case 'Pass':
		{
			if ( fPassed )
			{
				fMessageView->SetText("Game over.");
			} else {
				fPassed = true;
				
				// set up for own move
				fGo->UpdateValidPlacements( fGoView->fColor );
				fGoView->SetEnabled(true);
				fPassButton->SetEnabled(true);
				fMessageView->SetText("Opponent passed. Your move.");
			}
		}	break;
		
		case 'Quit':
		{
			fMessageView->SetText("Opponent left game.");
			fGoView->SetEnabled(false);
			fPassButton->SetEnabled(false);
		}	break;
			
		default:
			printf("Unknown message in GoWindow::HandleGoMessage()\n");
			msg->PrintToStream();
			break;
	}
}

void
GoWindow::GoEvent( int stone, XY xy )
{
	BAutolock lock(this);
	
	if ( stone == fGoView->fColor )
	{ // send event to network
		fPassed = false;
		
		MessageRef msg = GetMessageFromPool('Move');
		
		msg()->AddString(PR_NAME_KEYS, fRemoteAddress.c_str() );
		msg()->AddString("go_session", fSession.c_str() );
		msg()->AddInt32("x", xy.x );
		msg()->AddInt32("y", xy.y );
		
		gMuscle.SendMessageToSessions(msg);
		
		fGoView->SetEnabled(false);
		fPassButton->SetEnabled(false);
		fMessageView->SetText("Opponent's move..");
		UpdateScore();
	}
}

void
GoWindow::UpdateScore()
{
	char score[50];
	
	sprintf(score,"W: %d B: %d", fGo->WhiteScore(), fGo->BlackScore() );
	
	fScoreView->SetText(score);
}
